import gym
from gym.envs.toy_text.frozen_lake import generate_random_map
import numpy as np
from scipy.linalg import eig
import datetime as dt
import matplotlib.pyplot as plt
from tqdm import tqdm
import pandas as pd

import ray
import psutil

def flatten_mdp(policy, model):
    nstates = policy.shape[0]
    P_pi = np.zeros([nstates, nstates])  # transition probability matrix (s) to (s')
    R_pi = np.zeros([nstates])             # exp. reward from state (s) to any next state
    for s in range(nstates):
        for a in range(nactions):
            for p_, s_, r_, _ in model[s][a]:
                P_pi[s, s_] += policy[s,a] * p_   # transition probability (s) -> (s')
                Rsa = p_ * r_                     # exp. reward from (s,a) to any next state
                R_pi[s] += policy[s,a] * Rsa      # exp. reward from (s) to any next state
    assert np.allclose(P_pi.sum(axis=1),1)  # rows should sum to 1
    return P_pi, R_pi

def gen_data(env, policy, n_games):
    Data = []
    nactions = env.env.nA
    for i in range(n_games):
        data = []
        done=False
        state=env.reset()
        while not done:
            action = np.random.choice(nactions, p=policy[state,:])
            state_, reward, done, info = env.step(action)
            data.append((state,reward,state_))
            state = state_
        Data.append(data)
    return Data

def online_bootstrap(Data, Phi, gamma, alpha0, eta, B, freq=10):
    CIs = []
    Theta = Thetabar = np.zeros([B+1,Phi.shape[1]])
    n_games = len(Data)
    j=0

    for i in range(n_games):
        data = Data[i]
        for k in range(len(data)):
            j+=1
            alpha_t = alpha0 * j**(-eta)
            W = np.concatenate([[1], np.random.exponential(size=B)])
            state,reward,state_ = data[k]
            phi = Phi[state,:]
            phi_ = Phi[state_,:]
            At = np.outer(phi, (phi-gamma*phi_))
            bt = reward*phi
            Theta += alpha_t*np.diag(W) @ (bt[np.newaxis,:] - Theta@At.T)
            Thetabar = ((j-1)*Thetabar + Theta)/j
        if i % freq == 0:
            thetabar = Thetabar[0,:]
            ThetabarW = Thetabar[1:,:]
            SE = np.sqrt(np.diag(Phi @ np.cov(thetabar[np.newaxis,:] - ThetabarW, rowvar=False) @ Phi.T))
            CI = (Phi @ thetabar)[:,np.newaxis] + 1.96*np.outer(SE, [-1,1])
            CI = np.insert(CI,1,Phi@thetabar,axis=1)
            CIs.append(CI)
    return CIs

def offline_bootstrap(Data, Phi, gamma, alpha0, eta, B, freq=10):
    CIs = []
    Theta = Thetabar = np.zeros([B+1,Phi.shape[1]])
    n_games = len(Data)
    samples = [np.arange(n_games)] + [np.random.choice(n,size=n,replace=True) for _ in range(B)]
    js=np.zeros(B+1).astype(int)

    for i in range(n_games):
        for b in range(B+1):
            data = Data[samples[b][i]]
            for k in range(len(data)):
                js[b]+=1
                alpha_t = alpha0 * js[b]**(-eta)
                state,reward,state_ = data[k]
                phi = Phi[state,:]
                phi_ = Phi[state_,:]
                At = np.outer(phi, (phi-gamma*phi_))
                bt = reward*phi
                Theta[b,:] += alpha_t*(bt - At@Theta[b,:])
                Thetabar[b,:] = ((js[b]-1)*Thetabar[b,:] + Theta[b,:])/js[b]
        if i % freq == 0:
            thetabar = Thetabar[0,:]
            ThetabarW = Thetabar[1:,:]
            SE = np.sqrt(np.diag(Phi @ np.cov(thetabar[np.newaxis,:] - ThetabarW, rowvar=False) @ Phi.T))
            CI = (Phi @ thetabar)[:,np.newaxis] + 1.96*np.outer(SE, [-1,1])
            CI = np.insert(CI,1,Phi@thetabar,axis=1)
            CIs.append(CI)
    return CIs

@ray.remote
def run_bootstrap(seed, env, policy, n_games, Phi, gamma, alpha0, eta, B, freq=10):
    np.random.seed(seed)
    Data = gen_data(env, policy, n_games)
    return online_bootstrap(Data, Phi, gamma, alpha0, eta, B, freq)

if __name__ == "__main__":
    num_cpus = psutil.cpu_count(logical=False)
    ray.init(num_cpus=num_cpus, ignore_reinit_error=True)

    np.random.seed(2021)
    random_map = generate_random_map(size=8,p=0.8)
    env = gym.make('FrozenLake-v1', desc=random_map)

    nstates = env.env.nS
    nactions = env.env.nA
    P_MDP = env.env.P

    # generate feature matrix (row n corresponds to state n)
    gamma = 0.99
    p = 4
    Phi = np.random.random(size=(nstates, p))

    # create policy matrix from Q table
    Q_table = np.loadtxt('Q_fz.csv', delimiter=',')
    policy_Q = np.array([np.eye(nactions)[np.argmax(row),:] if row.sum() > 0 else np.ones(nactions)/nactions for row in Q_table])
    policy_random = np.ones([nstates, nactions]) / nactions  # 0.25 probability for each action

    epsilon = 0.2
    policy = (1-epsilon)*policy_Q + epsilon*policy_random

    P_pi, R_pi = flatten_mdp(policy, model=P_MDP)
    Abar = np.eye(nstates) - gamma*P_pi
    bbar = R_pi

    # convert to infinite horizon TPM
    for i in range(nstates):
        if P_pi[i,i] == 1.0:
            P_pi[i,i] = 0
            P_pi[i,0] = 1.0

    eigen = eig(P_pi,left=True)
    tmp = np.real(eigen[1][:,np.argmin(np.abs(eigen[0] - 1))])
    stat_dist = tmp / np.sum(tmp)
    D_pi = np.diag(stat_dist)

    Abar = Phi.T @ D_pi @ (np.eye(nstates) - gamma*P_pi) @ Phi
    bbar = Phi.T @ D_pi @ R_pi
    theta0 = np.linalg.pinv(Abar) @ bbar # analytical solution
    value0 = np.dot(Phi, theta0) # true value functions

    # params
    niter = 200
    n_games = int(2e3)
    alpha0 = 5
    eta = 2/3
    B = 200
    args = {'env':env, 'policy':policy, 'n_games':n_games, 'Phi':Phi, 'gamma':gamma, 'B':B, 'freq':10,
           }

    #widths
    Data = gen_data(env, policy, n_games)
    alpha0_list = [2,5,10]
    eta_list = [0.6,0.7,0.8]

    CIs_alpha0 = {alpha0: online_bootstrap(Data, Phi, gamma, alpha0, eta, B, freq=10) for alpha0 in tqdm(alpha0_list)}
    CIs_eta = {eta: online_bootstrap(Data, Phi, gamma, alpha0, eta, B, freq=10) for eta in tqdm(eta_list)}

    dim = 0
    CIs_width_alpha0 = pd.DataFrame({k: np.array([CI[dim,2] - CI[dim,0] for CI in v]) for k,v in CIs_alpha0.items()})
    CIs_width_eta = pd.DataFrame({k: np.array([CI[dim,2] - CI[dim,0] for CI in v]) for k,v in CIs_eta.items()})

    CIs_width_alpha0.set_index(np.arange(len(CIs_width_alpha0.index))*10, inplace=True)
    CIs_width_eta.set_index(np.arange(len(CIs_width_eta.index))*10, inplace=True)

    plt.rcParams['figure.figsize'] = (16,9)
    plt.rcParams['figure.facecolor'] = 'white'
    CIs_width_alpha0.loc[500:,:].plot(linewidth=3)
    plt.xlabel('Number of episodes', fontsize='xx-large')
    plt.ylabel('CI Width', fontsize='xx-large')
    plt.legend(fontsize='xx-large')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.savefig('frozenlake_sensitivity_width_alpha0.png', bbox_inches='tight')
    plt.close()

    plt.rcParams['figure.figsize'] = (16,9)
    plt.rcParams['figure.facecolor'] = 'white'
    CIs_width_eta.loc[500:,:].plot(linewidth=3)
    plt.xlabel('Number of episodes', fontsize='xx-large')
    plt.ylabel('CI Width', fontsize='xx-large')
    plt.legend(fontsize='xx-large')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.savefig('frozenlake_sensitivity_width_eta.png', bbox_inches='tight')
    plt.close()

    #coverage
    res_alpha0 = dict()
    for alpha0 in tqdm(alpha0_list):
        res_alpha0[alpha0] = ray.get([run_bootstrap.remote(seed, alpha0=alpha0, eta=eta, **args) for seed in range(niter)])

    res_eta = dict()
    for eta in tqdm(eta_list):
        res_eta[eta] = ray.get([run_bootstrap.remote(seed, alpha0=alpha0, eta=eta, **args) for seed in range(niter)])

    cov_alpha0 = pd.DataFrame({k: np.array([np.array([(value0[dim] > CIs[dim,0]) & (value0[dim] < CIs[dim,2]) for CIs in r]) for r in v]).mean(axis=0) for k,v in res_alpha0.items()})
    cov_eta = pd.DataFrame({k: np.array([np.array([(value0[dim] > CIs[dim,0]) & (value0[dim] < CIs[dim,2]) for CIs in r]) for r in v]).mean(axis=0) for k,v in res_eta.items()})

    x = np.arange(len(cov_alpha0.index))*10
    cov_alpha0.set_index(x, inplace=True)
    cov_eta.set_index(x, inplace=True)

    plt.rcParams['figure.figsize'] = (16,9)
    plt.rcParams['figure.facecolor'] = 'white'
    cov_alpha0.plot(linewidth=2)
    plt.axhline(0.95, linestyle='--', color='black', linewidth=3)
    plt.ylim((0,1))
    plt.xlabel('Number of episodes', fontsize='xx-large')
    plt.ylabel('CI Coverage ({} runs)'.format(niter), fontsize='xx-large')
    plt.legend(fontsize='xx-large')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.savefig('frozenlake_sensitivity_coverage_alpha0.png', bbox_inches='tight')
    plt.close()

    plt.rcParams['figure.figsize'] = (16,9)
    plt.rcParams['figure.facecolor'] = 'white'
    cov_eta.plot(linewidth=2)
    plt.axhline(0.95, linestyle='--', color='black', linewidth=3)
    plt.ylim((0,1))
    plt.xlabel('Number of episodes', fontsize='xx-large')
    plt.ylabel('CI Coverage ({} runs)'.format(niter), fontsize='xx-large')
    plt.legend(fontsize='xx-large')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.savefig('frozenlake_sensitivity_coverage_eta.png', bbox_inches='tight')
    plt.close()
